/***************************************************************************
*
*	File name : flashc_drv.c
*	Function  : flashc driver (C17801)
*	Revision  :
*		July. 27, 2007	Yakushiji		First revision
*
****************************************************************************/

// Include File
#include	"reg_801.h"
#include	"system.h"
#include	"flashc_api.h"

#define	CTL_STOP			0x0400
#define	CTL_STARTHOLD		0x0200
#define	CTL_STARTERASE		0x0100
#define	CTL_ERASE_CHIP		FLASHC_ERASE_CHIP
#define	CTL_ERASE_SECTOR	FLASHC_ERASE_SECTOR
#define	CTL_WRITE			0x0001

#ifndef	FOR_FLASHC_MINIMUM_API
void			FlashcInit(T_FLASHC_CFG*);
void			FlashcSetTTBR(unsigned long);
unsigned long	FlashcGetTTBR(void);
#endif
int				FlashcErase(unsigned char, unsigned long, unsigned long);
int				FlashcWrite(unsigned long, unsigned long, unsigned short*);
static void		Wait(unsigned long);

#ifndef	FOR_FLASHC_MINIMUM_API
/*************************************************************
	FlashcInit()

	description:
		Initialize of Flash controller.

	argument	: None
	return		: None

**************************************************************/
void FlashcInit(T_FLASHC_CFG *config){

	FlashcSetFlashWait(config->flashWait);

	if(config->TTBR != 0xFFFFFFFF){
		FlashcUnprotect();
		FlashcSetTTBR(config->TTBR);
		FlashcProtect();
	}

}

/*************************************************************
	FlashcSetTTBR()

	description:
		Set Trap table base address.

	argument	:
		TTBR : Trap table base address.

	return		: None

**************************************************************/
void FlashcSetTTBR(unsigned long TTBR){

	FLSHC_TTBRH = ((unsigned short *)&TTBR)[1];
	FLSHC_TTBRL = ((unsigned short *)&TTBR)[0];
}

/*************************************************************
	FlashcGetTTBR()

	description:
		Get Trap table base address.

	argument	: None

	return		:
		TTBR : Trap table base address.

**************************************************************/
unsigned long FlashcGetTTBR(void){

	unsigned long	TTBR;

	((unsigned short *)&TTBR)[1] = FLSHC_TTBRH;
	((unsigned short *)&TTBR)[0] = FLSHC_TTBRL;

	return TTBR;
}
#endif

/*************************************************************
	FlashcErase()

	description:
		Set Trap table base address.

	argument	:
		mode        : Chip Erase or Sector Erase
		startAddress: Erase start address
		endAddress  : Erase end address

	return		:
		ERROR  : Parameter error or erase error
		SUCCESS: Erase complete

**************************************************************/
int FlashcErase(unsigned char mode, unsigned long startAddress, unsigned long endAddress){

	int				ret = ERROR;
	unsigned long	wait1 = 0;
	unsigned long	wait2 = 0;
	unsigned long	address;

	// Check start and end address
	if((startAddress >= FLASHC_TOP_ADDRESS && startAddress <= FLASHC_END_ADDRESS) ||
	   (endAddress >= FLASHC_TOP_ADDRESS && endAddress <= FLASHC_END_ADDRESS) ||
	   (startAddress <= endAddress)){

		startAddress -= FLASHC_TOP_ADDRESS;
		endAddress   -= FLASHC_TOP_ADDRESS;

		startAddress = (startAddress >> 10) & 0x7F;
		endAddress   = (endAddress >> 10) & 0x7F;

		// Check erase type
		if(mode == CTL_ERASE_CHIP){
			wait1 = 10000;
			wait2 = 2;
		} else if(mode == CTL_ERASE_SECTOR){
			wait1 = 2000;
			wait2 = 1;
		}

		if(wait1 != 0){
			address = startAddress;

			while(1){

				ret = ERROR;

				if(mode == CTL_ERASE_SECTOR){
					FLSHC_ADDR = address;
				}

				// if Flash is idle then erase process
				if((FLSHC_CTL & FLSHC_CTL_STS) == 0){

					// Flash controller unprotect
					FlashcUnprotect();

					FLSHC_CTL = mode;

					// start erase
					FLSHC_CTL |= CTL_STARTERASE;

					// if idle status then error
					if((FLSHC_CTL & FLSHC_CTL_STS) == 0){
						// Clear Flash controller
						FLSHC_CTL = 0;

						// Flash controller protect
						FlashcProtect();

						break;
					}

					// wait 100msec or 20msec
					Wait(wait1);

					// start hold
					FLSHC_CTL |= CTL_STARTHOLD;

					// wait 20usec or 10usec
					Wait(wait2);

					// stop
					FLSHC_CTL |= CTL_STOP;

					// if busy status then error
					if(FLSHC_CTL & FLSHC_CTL_STS){
						// Clear Flash controller
						FLSHC_CTL = 0;

						// Flash controller protect
						FlashcProtect();

						break;
					}

					// Clear Flash controller
					FLSHC_CTL = 0;

					// Flash controller protect
					FlashcProtect();

					if(*(volatile unsigned short *)((address << 10) + FLASHC_TOP_ADDRESS) != 0xFFFF){
						break;
					}

					ret = SUCCESS;

					address++;

					if(mode == CTL_ERASE_CHIP || address > endAddress){
						break;
					}
				}
			}
		}
	}

	return ret;
}


/*************************************************************
	FlashcWrite()

	description:
		Writing to Flash.

	argument	:
		address : start write address
		size    : data size
		data    : data pointer

	return		:
		ERROR  : Parameter error or write error
		SUCCESS: Writing complete

**************************************************************/
int FlashcWrite(unsigned long address, unsigned long size, unsigned short *data){

	int				ret = ERROR;
	unsigned long	i;
	unsigned long	j;

	// Check address
	if(address >= FLASHC_TOP_ADDRESS && address <= FLASHC_END_ADDRESS && ((address & 0x00001) == 0)){

		// if Flash is idle then erase process
		if((FLSHC_CTL & FLSHC_CTL_STS) == 0){

			// Flash controller unprotect
			FlashcUnprotect();

			FLSHC_CTL = CTL_WRITE;

			for(i = 0, j = 0; i < size; i+=2, j++){

				*(unsigned short *)(address + i) = data[j];

				// if idle status then error
				if((FLSHC_CTL & FLSHC_CTL_STS) == 0){
					// Clear Flash controller
					FLSHC_CTL = 0;

					// Flash controller protect
					FlashcProtect();

					return ret;
				}

				// wait 20usec
				Wait(2);

				// start hold
				FLSHC_CTL |= CTL_STARTHOLD;

				// wait 10usec
				Wait(1);

				// stop
				FLSHC_CTL |= CTL_STOP;

				// if busy status then error
				if(FLSHC_CTL & FLSHC_CTL_STS){
					// Clear Flash controller
					FLSHC_CTL = 0;

					// Flash controller protect
					FlashcProtect();

					return ret;
				}
			}

			// Clear Flash controller
			FLSHC_CTL = 0;

			// Flash controller protect
			FlashcProtect();

			// Verify check
			for(i = 0, j = 0; i < size; i+=2, j++){
				if(*(volatile unsigned short *)(address + i) != data[j]){
					return ret;
				}
			}
			ret = SUCCESS;
		}
	}

	return ret;
}

/*************************************************************
	Wait()

	description:
		wait.

	argument	:
		wait    : wait value

	return		: None

**************************************************************/
static void Wait(unsigned long wait){

	int		i;

	// At first, user should execute the time of 'Wait(10)'. User must
	// adjust the number of 'nop' so that time from point A to point B
	// becomes nearest to 100 [usec].
#ifdef	DEBUG_WAIT
	// check point A
	EGPIO_P1_DATA |= EGPIO_P1_DATA_P11D;
#endif

	while (wait > 0){
		for(i = 0; i < 2; i++){
			asm("jpr  0");
			asm("jpr  0");
			asm("jpr  0");
			asm("jpr  0");
			asm("jpr  0");
			asm("jpr  0");
			asm("jpr  0");
			asm("jpr  0");
			asm("jpr  0");
			asm("jpr  0");
		}

		wait--;
	}

#ifdef	DEBUG_WAIT
	// check point B
	EGPIO_P1_DATA &= ~EGPIO_P1_DATA_P11D;
#endif
}
