/***************************************************************************
*
*	File name : remc_drv.c
*	Function  : REMC driver (C17801)
*	Revision  :
*		July. 30, 2007	Yakushiji		First revision
*
****************************************************************************/

// Include File
#include	"reg_801.h"
#include	"system.h"
#include	"remc_api.h"

static T_REMC_CFG		Config;			// save initialize value
static unsigned long	Fprescale;		// carrier frequency

static unsigned short	LeaderOn;		// Leader Code ON pulse wide
static unsigned short	LeaderOff;		// Leader Code OFF pulse wide
static unsigned short	Data0On;		// Data Code 0 ON pulse wide
static unsigned short	Data0Off;		// Data Code 0 OFF pulse wide
static unsigned short	Data1On;		// Data Code 1 ON pulse wide
static unsigned short	Data1Off;		// Data Code 1 OFF pulse wide
static unsigned short	StopOn;			// Stop bit ON pulse wide
static unsigned short	StopOff;		// Stop bit OFF pulse wide

#ifdef FOR_REMC_RECEIVE
static unsigned short	LeaderOn_Max;		// Leader Code ON pulse wide
static unsigned short	LeaderOn_Min;		// Leader Code ON pulse wide
static unsigned short	LeaderOff_Max;		// Leader Code OFF pulse wide
static unsigned short	LeaderOff_Min;		// Leader Code OFF pulse wide
static unsigned short	Data0On_Max;		// Data Code 0 ON pulse wide
static unsigned short	Data0On_Min;		// Data Code 0 ON pulse wide
static unsigned short	Data0Off_Max;		// Data Code 0 OFF pulse wide
static unsigned short	Data0Off_Min;		// Data Code 0 OFF pulse wide
static unsigned short	Data1On_Max;		// Data Code 1 ON pulse wide
static unsigned short	Data1On_Min;		// Data Code 1 ON pulse wide
static unsigned short	Data1Off_Max;		// Data Code 1 OFF pulse wide
static unsigned short	Data1Off_Min;		// Data Code 1 OFF pulse wide
static unsigned short	StopOn_Max;			// Stop bit ON pulse wide
static unsigned short	StopOn_Min;			// Stop bit ON pulse wide
static unsigned short	StopOff_Max;		// Stop bit OFF pulse wide
static unsigned short	StopOff_Min;		// Stop bit OFF pulse wide
#endif

static unsigned short	Average;		// Average error

static unsigned char	*Buffer;		// save buffer pointer
static unsigned short	BitCount;		// bit counter for data code
static unsigned short	ByteCount;		// byte counter for buffer
static unsigned char	BitData;		// transmit bit data
static void				(*InterruptMaskFunc)(unsigned char);

const static unsigned char	ShiftBit[]={0,1,2,3,4,5,6,7};	// shift bit

static unsigned short	State;			// Transmit/Receive state

static unsigned short	InterruptFactor;
static unsigned short	Capture;

// State
#define	STATE_IDLE		0
#define	STATE_DUMMY		1
#define	STATE_LEADER_H	2
#define	STATE_LEADER_L	3
#define	STATE_DATA_H	4
#define	STATE_DATA_L	5
#define	STATE_STOP_H	6
#define	STATE_STOP_L	7
#define	STATE_COMP		8

static void	Initialize(void);
static void InterruptMask(unsigned char);

/*************************************************************
	RemcInit()

	description:
		Initialize of Remote controller.

	argument	: None
	return		: None

**************************************************************/
void RemcInit(T_REMC_CFG *config){

	unsigned long	divide;
	unsigned long	countTime;
	unsigned short	TOn;
	unsigned short	TOff;


	// save configuration
	Config.mode			= config->mode;
	Config.systemClock	= config->systemClock;
	Config.divide		= config->divide;
	Config.carrierHigh	= config->carrierHigh;
	Config.carrierLow	= config->carrierLow;

	// Bit count of Data Code
	Config.dataCodeBit	= config->dataCodeBit;

	// ON pulse wide
	Config.T_ON			= config->T_ON;
	// OFF pulse wide
	Config.T_OFF		= config->T_OFF;

	Config.averageError	= config->averageError;

	Config.leaderON		= config->leaderON;
	Config.leaderOFF	= config->leaderOFF;
	Config.data0ON		= config->data0ON;
	Config.data0OFF		= config->data0OFF;
	Config.data1ON		= config->data1ON;
	Config.data1OFF		= config->data1OFF;
	Config.stopON		= config->stopON;
	Config.stopOFF		= config->stopOFF;

	// Interrupt mask function
	InterruptMaskFunc	= config->intMaskFunc;

	// Change set divide value
	if(Config.divide == REMC_PSDIV_32){
		divide = 32;
	} else if(Config.divide == REMC_PSDIV_24){
		divide = 24;
	} else if(Config.divide == REMC_PSDIV_16){
		divide = 16;
	} else {
		divide = 12;
	}

	// e.g System Clock 48MHz
	// Fprescale = 48000000 / 24 = 2000000 = 2MHz
	Fprescale = config->systemClock / divide;

	// countTIme = (32 / 2MHz) * 1000000 * 10 = 16us * 10
	countTime = 320000000 / Fprescale;

	// e.g NEC Format T_ON = 560us T_OFF = 560us
	// Txx = 560us / 16us * 10
	TOn  = (Config.T_ON * 100) / countTime;
	TOff = (Config.T_OFF * 100) / countTime;

	Average = ((Config.averageError * 10) / countTime) + 1;

	// xxxxOn = envelope count vale
	LeaderOn  = (Config.leaderON  * TOn)  / 10;
	LeaderOff = (Config.leaderOFF * TOff) / 10;
	Data0On   = (Config.data0ON   * TOn)  / 10;
	Data0Off  = (Config.data0OFF  * TOff) / 10;
	Data1On   = (Config.data1ON   * TOn)  / 10;
	Data1Off  = (Config.data1OFF  * TOff) / 10;
	StopOn    = (Config.stopON    * TOn)  / 10;
	StopOff   = (Config.stopOFF   * TOff) / 10;

#ifdef FOR_REMC_RECEIVE
	LeaderOn_Max	= LeaderOn + Average;
	LeaderOn_Min	= LeaderOn - Average;
	LeaderOff_Max	= LeaderOff + Average;
	LeaderOff_Min	= LeaderOff - Average;
	Data0On_Max		= Data0On + Average;
	Data0On_Min		= Data0On - Average;
	Data0Off_Max	= Data0Off + Average;
	Data0Off_Min	= Data0Off - Average;
	Data1On_Max		= Data1On + Average;
	Data1On_Min		= Data1On - Average;
	Data1Off_Max	= Data1Off + Average;
	Data1Off_Min	= Data1Off - Average;
	StopOn_Max		= StopOn + Average;
	StopOn_Min		= StopOn - Average;
	StopOff_Max		= StopOff + Average;
	StopOff_Min		= StopOff - Average;
#endif

	State = STATE_IDLE;

	Initialize();
}

/*************************************************************
	Initialize()

	description:
		Initialize of remote controller.

	argument	: None
	return		: None

**************************************************************/
static void Initialize(void){

	// stop module
	RemcStopREMC();
	RemcStopPrescale();

	RemcSetPrescale(Config.divide);
	RemcStartPrescale();

	// clear interrupt factor
	InterruptFactor = 0;

	// Transmit mode
	if(Config.mode == REMC_TRANSMIT){

		// set carrier
		RemcSetCarrier(Config.carrierHigh, Config.carrierLow);

		// RemcSetMode(mode);
		// RemcEnableUnderflow();
		// Clear Interrupt
		REMC_CNF = (Config.mode | REMC_CNF_UIE | 0x0070);

	// Receive mode
	} else {

		// set carrier
		RemcSetCarrier(0, 0);
		// set envelope loop register
		RemcSetTransmit(0xFFFF);

		// RemcSetMode(mode);
		// RemcEnableFalling();
		// RemcEnableRising();
		// Clear Interrupt
		REMC_CNF = (Config.mode | REMC_CNF_FIE | REMC_CNF_RIE | 0x0070);
	}
}

#ifdef	FOR_REMC_TRANSFER
/*************************************************************
	RemcTransmit()

	description:
		Send data.

	argument	:
		buffer	: pointer of send data
	return		:
		SUCCESS    : Complete
		PROCESSING : Transmit processing

**************************************************************/
int RemcTransmit(unsigned char *buffer){

	int ret = PROCESSING;

	// if idle then transmit mode
	if(State == STATE_IDLE){

		Config.mode = REMC_TRANSMIT;

		// save buffer pointer
		Buffer = buffer;
		// clear counter
		BitCount = 0;
		ByteCount = 0;

		// initialize hardware
		Initialize();

		// set leader code
		RemcSetTransmit(LeaderOn);

		// next state
		State = STATE_LEADER_L;

		// staer REMC
		RemcStartREMC();

	// if complete then transmit end
	} else if(State == STATE_COMP){
		State = STATE_IDLE;
		ret = SUCCESS;
	}

	return ret;
}
#endif

#ifdef	FOR_REMC_RECEIVE
/*************************************************************
	RemcReceive()

	description:
		Receive data.

	argument	:
		buffer	: pointer of receive data
	return		:
		SUCCESS    : Complete
		PROCESSING : Receive processing
		ERROR      : Receive error

**************************************************************/
int RemcReceive(unsigned char *buffer){

	int	ret = PROCESSING;

	// if idle then receive mode
	if(State == STATE_IDLE){
		Config.mode = REMC_RECEIVE;

		// save buffer pointer
		Buffer = buffer;
		// clear counter
		BitCount = 0;
		ByteCount = 0;

		// initialize hardware
		Initialize();

		// state is leader code
		State = STATE_LEADER_H;

		// start REMC
		RemcStartREMC();

	// if complete then end receive mode
	} else if(State == STATE_COMP){
		State = STATE_IDLE;
		ret = SUCCESS;
	}

	return ret;
}
#endif

/*************************************************************
	RemcStop()

	description:
		Stop REMC.

	argument	: NONE
	return		: NONE

**************************************************************/
void RemcStop(void){

	// stop module
	RemcStopREMC();
	RemcStopPrescale();

	// Clear Interrupt
	REMC_CNF = 0x0070;

	// clear interrupt factor
	InterruptFactor = 0;

	State = STATE_IDLE;
}

/*************************************************************
	RemcInterruptHandler()

	description:
		REMC interrupt handler.

	argument	: None
	return		: None

**************************************************************/
void RemcInterruptHandler(void){

	// save interrupt factor
	InterruptFactor |= (REMC_CNF & (REMC_CNF_RIF | REMC_CNF_FIF | REMC_CNF_UIF));

	// if receive factor then get receive data
	if(InterruptFactor & (REMC_CNF_RIF | REMC_CNF_FIF)){
		// Get capture pulse wide
		Capture = RemcGetReceive();
	}

	// clear interrupt factor
	REMC_CNF |= REMC_CNF;
}

/*************************************************************
	RemcInterruptProcess()

	description:
		REMC interrupt process.

	argument	: None
	return		: None

**************************************************************/
void RemcInterruptProcess(void){

	if(InterruptFactor == 0){
		return;
	}

#ifdef	FOR_REMC_TRANSFER
	// if transmit mode then transfer processing
	if(Config.mode == REMC_TRANSMIT){
		// check interrupt
		if(InterruptFactor & REMC_CNF_UIF){

			switch(State){
			case STATE_LEADER_L:
				RemcSetTransmit(LeaderOff);
				State++;
				break;
			case STATE_DATA_H:
				BitData = (Buffer[ByteCount] >> ShiftBit[BitCount & 0x7]) & 0x01;

				if(BitData){
					RemcSetTransmit(Data1On);
				} else {
					RemcSetTransmit(Data0On);
				}
				State++;
				break;

			case STATE_DATA_L:
				if(BitData){
					RemcSetTransmit(Data1Off);
				} else {
					RemcSetTransmit(Data0Off);
				}

				BitCount++;

				if((BitCount & 0x07) == 0){
					ByteCount++;
				}

				if(BitCount >= Config.dataCodeBit){
					State++;
				} else {
					State = STATE_DATA_H;
				}

				break;
			case STATE_STOP_H:
				RemcSetTransmit(StopOn);
				State++;
				break;
			case STATE_STOP_L:
				RemcSetTransmit(StopOff);
				State = STATE_DUMMY;
				break;
			case STATE_DUMMY:
			default:
				State = STATE_COMP;
				RemcStopREMC();
				break;
			}
		}

		InterruptFactor &= ~(REMC_CNF_UIF | REMC_CNF_RIF | REMC_CNF_FIF);

	// if not transmit mode then receive processing
	} else {
#endif
#ifdef	FOR_REMC_RECEIVE

		// Rising
		if(InterruptFactor & REMC_CNF_RIF){
			switch(State){
			case STATE_DUMMY:
			case STATE_LEADER_H:
				if(Capture <= LeaderOn_Max && Capture >= LeaderOn_Min){
					State = STATE_LEADER_L;
				} else {
					State = STATE_DUMMY;
				}
				break;
			case STATE_DATA_H:
				if(Capture <= Data0On_Max && Capture >= Data0On_Min){
					State++;
				} else if(Capture <= Data1On_Max && Capture >= Data1On_Min){
					State++;
				} else {
					State = STATE_DUMMY;
				}
				break;
			case STATE_STOP_H:
				if(Capture <= StopOn_Max && Capture >= StopOn_Min){
					State = STATE_COMP;
					RemcStopREMC();
				}
				break;
			default:
				break;
			}

			InterruptFactor &= ~(REMC_CNF_RIF | REMC_CNF_UIF);
		}
		// Falling
		if(InterruptFactor & REMC_CNF_FIF){
			switch(State){
			case STATE_DUMMY:
				State++;
				break;
			case STATE_LEADER_L:
				if(Capture <= LeaderOff_Max && Capture >= LeaderOff_Min){
					State++;
					// if Receive Leader Code then reset BitCount and ByteCount
					BitCount = 0;
					ByteCount = 0;
				} else {
					State = STATE_DUMMY;
				}
				break;
			case STATE_DATA_L:
				if(Capture <= Data0Off_Max && Capture >= Data0Off_Min){
					Buffer[ByteCount] &= ~(1 << ShiftBit[BitCount & 0x7]);
					BitCount++;

					if((BitCount & 0x07) == 0){
						ByteCount++;
					}

					if(BitCount >= Config.dataCodeBit){
						State++;
					} else {
						State = STATE_DATA_H;
					}

				} else if(Capture <= Data1Off_Max && Capture >= Data1Off_Min){
					Buffer[ByteCount] |= (1 << ShiftBit[BitCount & 0x7]);
					BitCount++;

					if((BitCount & 0x07) == 0){
						ByteCount++;
					}

					// if receive all bit then next state
					if(BitCount >= Config.dataCodeBit){
						State++;
					} else {
						State = STATE_DATA_H;
					}

				} else {
					State = STATE_DUMMY;
				}
				break;
			case STATE_STOP_L:
			default:
				State = STATE_DUMMY;
				break;
			}

			InterruptFactor &= ~(REMC_CNF_FIF | REMC_CNF_UIF);
		}
#endif
#ifdef	FOR_REMC_TRANSFER
	}
#endif
}

/*************************************************************
	RemcSetPrescale()

	description:
		Set prescale divider ratio.

	argument	:
		divider	: Prescale divider ratio
	return		:
		SUCCESS	: Success complete
		ERROR	: Ignore set value

**************************************************************/
int RemcSetPrescale(unsigned char divider){

	int		ret = ERROR;

	// if Prescaler clock OFF then change value
	if((REMC_PSC & REMC_PSC_PSON) == 0){

		REMC_PSC = (unsigned short)divider;

		ret = SUCCESS;
	}

	return ret;
}

/*************************************************************
	RemcSetCarrier()

	description:
		Set carrier High/Low value.

	argument	:
		high	: carrier high value
		low		: carrier low value
	return		:
		divider	: Prescale divider ratio

**************************************************************/
void RemcSetCarrier(unsigned char high, unsigned char low){

	unsigned short	value;


	((unsigned char *)&value)[1] = high;
	((unsigned char *)&value)[0] = low;

	REMC_CARL = value;
}

/*************************************************************
	InterruptMask()

	description:
		Interrupt mask.

	argument	:
		mask	: REMC_INT_ENABLE
				: REMC_INT_DISABLE
	return		: None

**************************************************************/
static void InterruptMask(unsigned char mask){

	if(InterruptMaskFunc != NULL){
		InterruptMaskFunc(mask);
	} else {
		if(mask == REMC_INT_ENABLE){
			CpuIntEnable();
		} else {
			CpuIntDisable();
		}
	}
}

